# Copyright 2022 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import os.path
import subprocess
from email.utils import parseaddr
from urllib.parse import urlparse

from charmhelpers.core import hookenv, host, templating
from charms.launchpad.payload import config_file_path
from charms.launchpad.payload import get_service_config as get_payload_config
from charms.reactive import endpoint_from_flag
from ols import base


def oopses_dir():
    return os.path.join(base.base_dir(), "oopses")


def scripts_dir():
    return os.path.join(base.base_dir(), "scripts")


def secrets_dir():
    return os.path.join(base.base_dir(), "secrets")


def var_dir():
    return os.path.join(base.base_dir(), "var")


def change_shell(user, shell):
    if (
        subprocess.run(
            ["getent", "passwd", user],
            capture_output=True,
            check=True,
            text=True,
        )
        .stdout.splitlines()[0]
        .split(":")[6]
        != shell
    ):
        subprocess.run(["chsh", "-s", shell, user], check=True)


def ensure_lp_directories():
    for dirpath in oopses_dir(), var_dir():
        host.mkdir(dirpath, group=base.user(), perms=0o775)
    host.mkdir(secrets_dir(), group=base.user(), perms=0o750)


def _get_first_rabbitmq_hostname(rabbitmq):
    for conversation in rabbitmq.conversations():
        for relation_id in conversation.relation_ids:
            for unit in hookenv.related_units(relation_id):
                return hookenv.relation_get(
                    "private-address", unit, relation_id
                )


def get_service_config():
    config = get_payload_config()
    config.update(
        {
            "oopses_dir": oopses_dir(),
            # Used by some templates.
            "parseaddr": parseaddr,
            "scripts_dir": scripts_dir(),
            "secrets_dir": secrets_dir(),
            "var_dir": var_dir(),
        }
    )

    # oops-datedir2amqp is used in crontabs and needs broken-out RabbitMQ
    # credentials.  This isn't ideal because it doesn't support high
    # availability, but we can tolerate that for fallback OOPS publishing
    # for now.
    if config["rabbitmq_broker_urls"]:
        rabbitmq_url = urlparse(config["rabbitmq_broker_urls"].split()[0])
        config.update(
            {
                "rabbitmq_host": rabbitmq_url.hostname,
                "rabbitmq_password": rabbitmq_url.password,
                "rabbitmq_username": rabbitmq_url.username,
                "rabbitmq_vhost": rabbitmq_url.path.lstrip("/"),
            }
        )
    else:
        rabbitmq = endpoint_from_flag("rabbitmq.available")
        hostname = _get_first_rabbitmq_hostname(rabbitmq)
        if hostname is not None:
            config.update(
                {
                    "rabbitmq_host": hostname,
                    "rabbitmq_password": rabbitmq.password(),
                    "rabbitmq_username": rabbitmq.username(),
                    "rabbitmq_vhost": rabbitmq.vhost(),
                }
            )

    return config


def lazr_config_files():
    return [
        config_file_path("launchpad-base-lazr.conf"),
        config_file_path("launchpad-base-secrets-lazr.conf", secret=True),
    ]


def configure_email(config, directory_name):
    mail_configure_path = config_file_path(
        f"{directory_name}/mail-configure.zcml"
    )
    if config["send_email"]:
        hookenv.log("Writing email configuration.")
        templating.render(
            "mail-configure.zcml",
            mail_configure_path,
            config,
            owner="root",
            group=base.user(),
            perms=0o444,
        )
    elif os.path.exists(mail_configure_path):
        hookenv.log("Removing email configuration.")
        os.unlink(mail_configure_path)


def configure_rsync(config, template, name):
    hookenv.log("Writing rsync configuration.")
    rsync_path = os.path.join("/etc/rsync-juju.d", name)
    if config["log_hosts_allow"]:
        templating.render(template, rsync_path, config, perms=0o644)
    elif os.path.exists(rsync_path):
        os.unlink(rsync_path)
