# Copyright 2019-2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
from collections import OrderedDict
import errno
import os
import shutil
import subprocess

from charmhelpers.core import hookenv, host, unitdata
from charmhelpers.payload import archive
import distro


# Layer configs #


def _layer_options():
    # import here, bc not available during unittest
    from charms import layer

    return layer.options().get('ols', {})


def service_name():
    return _layer_options().get('service_name')


def user():
    return _layer_options().get('user', 'ubuntu')


def tarball_payload():
    return _layer_options().get('tarball_payload', True)


def tarball_payload_name():
    """Return the name of the file in the object storage.

    By default is the same as service_name, but in multi-charm services
    it can be defined in layer.yaml to avoid having multiple copies of
    the same tarball with different names.
    """
    return _layer_options().get('tarball_payload_name', service_name())


def symlink_switch_payload():
    return _layer_options().get('symlink_switch_payload', True)


def python_bin():
    return _layer_options().get('python_bin', 'python3')


def use_system_site_packages():
    return _layer_options().get('use_system_site_packages', False)


# derived configs #


def root_dir():
    return '/srv'


def base_dir():
    basedir = "{}/{}".format(root_dir(), service_name())
    return basedir


def etc_dir():
    return '{}/etc'.format(base_dir())


def payloads_dir():
    return '{}/payloads'.format(base_dir())


def code_dir():
    return '{}/code'.format(base_dir())


def logs_dir():
    return '{}/logs'.format(base_dir())


def readonly_dirs():
    return [os.path.dirname(base_dir()), base_dir(), etc_dir(), payloads_dir()]


def writable_dirs():
    return [logs_dir()]


def service_config_path():
    return '{}/{}'.format(etc_dir(), service_config_filename())


def service_config_filename():
    return _layer_options().get('config_filename', 'settings.py')


def config_files():
    files = []
    files.append(service_config_path())
    return files


def service_paths():
    """Return default file paths for various things."""
    name = service_name()
    return {
        'config': service_config_path(),
        'working_dir': code_dir(),
        'log': '/var/log/{}.log'.format(name),
        'env': etc_dir() + '/{}.env'.format(name),
        'systemd': '/lib/systemd/system/{}.service'.format(name),
        'syslog': '/etc/rsyslog.d/22-{}.conf'.format(name),
        'logrotate': '/etc/logrotate.d/{}'.format(name),
    }


def payload_name():
    return (
        '{}.tar.gz'.format(tarball_payload_name())
        if tarball_payload()
        else service_name()
    )


def payload_dir():
    return os.path.join(hookenv.charm_dir(), 'files')


def version_info_path():
    return os.path.join(code_dir(), 'version-info.txt')


def env_vars():
    config = hookenv.config()
    env = OrderedDict()
    env['SERVICE_CONFIG'] = service_config_path()
    env['SERVICE_ENVIRONMENT'] = config['environment']
    env['SERVICE_DIR'] = code_dir()
    env['JUJU_MODEL'] = os.environ.get(
        'JUJU_MODEL_NAME', os.environ.get('JUJU_ENV_NAME')
    )
    env['JUJU_UNIT'] = os.environ.get('JUJU_UNIT_NAME')
    env['JUJU_MACHINE'] = os.environ.get('JUJU_MACHINE_ID')

    statsd_dsn = config.get('statsd_dsn')
    if statsd_dsn:
        env['STATSD_DSN'] = statsd_dsn
    sentry_dsn = config.get('sentry_dsn')
    if sentry_dsn:
        env['SENTRY_DSN'] = sentry_dsn

    kv = unitdata.kv()
    trusted_networks = config['trusted_networks'].split()
    extra_trusted_networks = kv.getrange('extra_trusted_networks.').values()
    # Flatten list of lists.
    extra_trusted_networks = [
        item for sublist in extra_trusted_networks for item in sublist
    ]
    env['TALISKER_NETWORKS'] = ' '.join(
        sorted(set(trusted_networks + extra_trusted_networks))
    )
    env['TALISKER_SOFT_REQUEST_TIMEOUT'] = config.get('soft_request_timeout')
    env['TALISKER_SLOWQUERY_THRESHOLD'] = config.get('slow_query_timeout')

    debuglog = config.get('debuglog')
    if debuglog:
        env['DEBUGLOG'] = debuglog

    extra_string = config.get('env_extra', '{}')
    try:
        extra = ast.literal_eval(extra_string)
    except ValueError:
        pass
    else:
        env.update(extra)

    return env


def set_extra_trusted_networks(key, networks):
    hookenv.log('Setting extra trusted networks: {}={}'.format(key, networks))
    kv = unitdata.kv()
    kv.update({key: networks}, prefix='extra_trusted_networks.')


def reset_extra_trusted_networks(key):
    hookenv.log('Removing extra trusted networks: {}'.format(key))
    kv = unitdata.kv()
    kv.unsetrange([key], prefix='extra_trusted_networks.')


# Utilities #


def ensure_directories(readonly, writable, user):
    hookenv.log('Creating service user')
    # ensure user and group
    host.add_group(user)
    host.adduser(user, primary_group=user)
    # ensure RO dirs
    hookenv.log('Creating service directories')
    for dirpath in readonly:
        host.mkdir(dirpath, group=user, perms=0o755)
    # ensure RW dirs
    for dirpath in writable:
        host.mkdir(dirpath, group=user, perms=0o775)


class PayloadError(Exception):
    pass


def get_swift_options(config):
    return {
        'username': config['swift_username'],
        'password': config['swift_password'],
        'auth_url': config['swift_auth_url'],
        'region_name': config['swift_region_name'],
        'tenant_name': config['swift_tenant_name'],
        'container_name': config['swift_container_name'],
        'storage_url': config['swift_storage_url'],
        'auth_version': config['swift_auth_version'],
        'user_domain_name': config['swift_user_domain_name'],
        'project_domain_name': config['swift_project_domain_name'],
    }


def swift_base_cmd(**swift_options):
    command = [
        'swift',
        '--os-username=' + swift_options['username'],
        '--os-password=' + swift_options['password'],
        '--os-auth-url=' + swift_options['auth_url'],
        '--os-region-name=' + swift_options['region_name'],
        '--os-tenant-name=' + swift_options['tenant_name'],
    ]
    # only add the domain and auth-version if auth-version=3
    if swift_options['auth_version'] == '3':
        command.extend(
            [
                '--auth-version=' + swift_options['auth_version'],
                '--os-user-domain-name=' + swift_options['user_domain_name'],
                '--os-project-domain-name=' + swift_options['project_domain_name'],
            ]
        )
    # if storage_url is defined, use it. This happens when the container is being
    # shared by other user
    if swift_options['storage_url']:
        command.append('--os-storage-url=' + swift_options['storage_url'])
    return command


def swift_fetch(source, target, **swift_options):
    from charms.layer.basic import apt_install

    container_name = swift_options['container_name']
    if not container_name:
        raise PayloadError('swift_container_name unset, so cannot fetch payload')
    if swift_options['username']:
        apt_install(['python3-swiftclient'])
        cmd = swift_base_cmd(**swift_options) + [
            'download',
            '--output=' + target,
            container_name,
            source,
        ]
    else:
        storage_url = swift_options['storage_url']
        if not storage_url:
            raise PayloadError(
                'swift_storage_url unset, so cannot fetch payload anonymously'
            )
        cmd = ['wget', '-O', target, '%s/%s/%s' % (storage_url, container_name, source)]
    subprocess.check_call(cmd)


def fetch_payload():
    config = hookenv.config()
    if not tarball_payload():
        raise PayloadError('Can only fetch tarball payloads')
    desired_build_label = config['build_label']
    if not desired_build_label:
        raise PayloadError('build_label unset, so cannot fetch payload')
    archive_path = os.path.join(payloads_dir(), '{}.tar.gz'.format(desired_build_label))
    object_name = os.path.join(
        desired_build_label, '{}.tar.gz'.format(tarball_payload_name())
    )
    swift_options = get_swift_options(config)
    if swift_options['container_name']:
        swift_fetch(
            # XXX: this {name}-builds parent directory is kept only for backwards
            # compatibility as buckets/containers can be used as a prefix.
            os.path.join('{}-builds'.format(tarball_payload_name()), object_name),
            archive_path,
            **swift_options
        )
    else:
        resource_path = hookenv.resource_get(service_name())
        if resource_path and os.path.getsize(resource_path):
            shutil.copy2(resource_path, archive_path)
            os.chmod(archive_path, 0o644)
        else:
            raise PayloadError('No payload available')
    return archive_path


def unlink_force(path):
    """Unlink path, without worrying about whether it exists."""
    try:
        os.unlink(path)
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise


def symlink_force(source, link_name):
    """Create symlink link_name -> source, even if link_name exists."""
    unlink_force(link_name)
    os.symlink(source, link_name)


def prune_payloads(keep_labels):
    for entry in os.listdir(payloads_dir()):
        entry_label = entry.split('.')[0]
        if entry_label in keep_labels:
            continue
        entry_path = os.path.join(payloads_dir(), entry)
        if os.path.isfile(entry_path) and entry.endswith('.tar.gz'):
            hookenv.log('Purging old payload tarball in {}'.format(entry_path))
            os.unlink(entry_path)
        elif os.path.isdir(entry_path):
            hookenv.log('Purging old build in {}'.format(entry_path))
            shutil.rmtree(entry_path)


def install_payload(payload, dest, owner):
    if symlink_switch_payload():
        config = hookenv.config()
        current_build_label = None
        if os.path.islink(dest):
            current_build_label = os.path.basename(os.path.realpath(dest))
        desired_build_label = config['build_label']
        if not desired_build_label:
            raise PayloadError('build_label unset, so cannot deploy code')
        if current_build_label == desired_build_label:
            hookenv.log('Build {} already deployed'.format(desired_build_label))
        # XXX: include build_label in the payload name to allow deploying via upgrade
        # charm if necessary, while still keeping track of deployed artifacts by
        # inspecting the config values or filesystem. This will result in the payload
        # being located at ${CHARM_DIR}/files/${GITHASH}/${SERVICE_NAME}.tar.gz
        payload = '{}/{}'.format(desired_build_label, payload)
        target_dir = os.path.join(payloads_dir(), desired_build_label)
    else:
        target_dir = dest

    payload_path = os.path.join(payload_dir(), payload)
    if not os.path.exists(target_dir):
        if not os.path.exists(payload_path):
            try:
                # XXX: this overrides the above payload_path value, now it points to
                # the download location
                payload_path = fetch_payload()
            except PayloadError as e:
                hookenv.status_set('blocked', str(e))
                raise

        # Support both tarball and raw payloads
        if os.path.isdir(payload_path):
            hookenv.log('Copying charm payload to {}'.format(target_dir))
            shutil.copytree(payload_path, target_dir, symlinks=True)
        else:
            hookenv.log('Extracting charm payload to {}'.format(target_dir))
            archive.extract_tarfile(payload_path, target_dir)
    else:
        hookenv.log("Payload already extracted in a previous run")

    venv_dir = os.path.join(target_dir, 'env')
    wheels_dir = os.path.join(target_dir, 'wheels')
    if os.path.exists(venv_dir) and not os.path.exists(wheels_dir):
        # there is a venv already and no wheels dir, relocate it!
        hookenv.log('Relocating virtualenv')
        subprocess.check_call(
            [
                'python2',
                '{}/scripts/virtualenv_tools.py'.format(hookenv.charm_dir()),
                '--update-path',
                'auto',
                venv_dir,
            ]
        )
    elif os.path.exists(wheels_dir):
        # there is a wheels directory and no env, let's create it
        create_virtualenv(wheels_dir, target_dir, python_bin())

    # code is read-only, only root has write access.
    host.chownr(target_dir, 'root', 'root', chowntopdir=True)

    if symlink_switch_payload():
        symlink_force(os.path.relpath(target_dir, os.path.dirname(dest)), dest)
        if current_build_label != desired_build_label:
            prune_payloads(keep_labels=[desired_build_label, current_build_label])


def create_virtualenv(wheels_dir, codedir, python_exe):
    env_dir = os.path.join(codedir, 'env')
    if os.path.exists(env_dir):
        hookenv.log('Virtualenv already exists, not creating it.')
    elif int(distro.major_version()) >= 22:
        # for all releases newer then 2022.04, we will use the
        # python-venv module
        create_venv(wheels_dir, codedir, python_exe, env_dir)
    else:
        # old behaviour
        hookenv.log('Installing python packages into the venv')

        cmd = ['virtualenv']

        if use_system_site_packages():
            hookenv.log('Enabling system site-packages in venv')
            cmd.append('--system-site-packages')

        if distro.linux_distribution()[-1] != 'trusty':
            cmd.append('--no-download')
            cmd.append('--extra-search-dir={}'.format(wheels_dir))

        subprocess.check_call(
            cmd
            + ['-p', python_exe, '--extra-search-dir={}'.format(wheels_dir), env_dir]
        )

    requirements = os.path.join(codedir, 'requirements.txt')

    hookenv.log('Installing from local wheelhouse.')
    pip_args = [
        '%s/bin/pip' % env_dir,
        'install',
        '--no-index',
        '--find-links={}'.format(wheels_dir),
        '-r',
        requirements,
    ]
    # The payload may have a setup.py.  If so, tell pip to install the code
    # directory as well, so that any entry points configured by the payload
    # will have corresponding scripts built for them.  We install this in
    # editable mode so that we don't end up with a superfluous (and possibly
    # incomplete, if the 'packages' parameter to setup is incorrect) copy of
    # the code under site-packages.
    if os.path.exists(os.path.join(codedir, 'setup.py')):
        pip_args.extend(['-e', codedir])
    subprocess.check_call(pip_args)


def create_venv(wheels_dir, codedir, python_exe, env_dir):
    # Install virtualenvs based on the python3 venv module
    # it only needs to be installed
    from charms.layer.basic import apt_install

    hookenv.log('Use python3 venv for virtualenv creation')

    apt_install(['python3-venv'])

    cmd = ['python3', '-m', 'venv']
    if use_system_site_packages():
        hookenv.log('Enabling system site-packages in venv')
        cmd.append('--system-site-packages')

    subprocess.check_call(cmd + [env_dir])


def render_template(template, context):
    from jinja2 import Template

    tmpl = Template(template, lstrip_blocks=True, trim_blocks=True)
    return tmpl.render(**context)


def check_port(key, new_port):
    """Open/close a port and keep track of it based on key:"""
    unitdata_key = '{}.port'.format(key)
    kv = unitdata.kv()
    existing_port = kv.get(unitdata_key)
    if existing_port != new_port:
        hookenv.open_port(new_port)
        if existing_port:  # Dont try to close non existing ports
            hookenv.close_port(existing_port)
        kv.set(unitdata_key, new_port)
