# Copyright 2019-2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import yaml
import subprocess
import traceback


from charmhelpers.core import hookenv, host
from charms import layer

from ols import base


class MainDatabaseNotAvailable(Exception):
    pass


def db_relations(only_admin=False):
    databases = layer.options().get('ols-pg', {}).get('databases', {})
    config_databases = yaml.safe_load(hookenv.config().get('databases', ''))
    for db_name, db_params in (config_databases or {}).items():
        databases[db_name].update(db_params)

    with open('metadata.yaml') as metadata_file:
        metadata = yaml.safe_load(metadata_file.read())
    requires = metadata.get('requires', {})
    relations = {}

    for relation, rel_type in requires.items():
        if rel_type['interface'] == 'pgsql':
            if relation.endswith('-admin'):
                name = relation.replace('-admin', '')
            elif only_admin:
                continue
            else:
                name = relation
            cfg = databases.get(name)
            if cfg:
                relations[relation] = databases.get(name, {})
    return relations


def write_admin_config(pgsql, filename='db-admin.yaml'):
    config_path = os.path.join(base.etc_dir(), filename)
    db_config = {}
    if pgsql.master:
        conn_info = {
            'host': pgsql.master.host,
            'user': pgsql.master.user,
            'port': pgsql.master.port,
            'password': pgsql.master.password,
            'dbname': hookenv.config()['db_name'],
        }
        db_config['master'] = conn_info
    if pgsql.standbys:
        standbys = []
        for s in pgsql.standbys:
            standbys.append(
                {'host': s.host, 'user': s.user, 'port': s.port, 'password': s.password}
            )
        db_config['standbys'] = standbys
    host.write_file(config_path, yaml.dump(db_config).encode('utf-8'))


def run_migration(pgsql, action=True, migrations='migrations'):
    schema_script = os.path.abspath('{}/env/bin/lp-migrate'.format(base.code_dir()))
    # allow the leaf layer to override which lp-migrate version to use
    if not os.path.exists(schema_script):
        schema_script = "lp-migrate"
    migrate_cmd = [schema_script, pgsql.master.uri, migrations]

    output = 'Running {} for {}@{}/{} from {}:\n'.format(
        schema_script,
        pgsql.master.user,
        pgsql.master.host,
        pgsql.master.dbname,
        migrations,
    )

    try:
        p = subprocess.Popen(
            migrate_cmd,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            close_fds=True,
            cwd=base.code_dir(),
        )
        cmd_output, _ = p.communicate()
    except Exception:
        if not action:
            raise
        return False, output + traceback.format_exc()

    output += cmd_output.decode('utf-8')
    return p.returncode == 0, output


def get_db_uris(pgsql):
    if pgsql.standbys:
        # avoid "" or None appearing in the list
        # Return a sorted list for charm configuration stability
        slave_uris = sorted([s.uri for s in pgsql.standbys if s.uri])
    else:
        slave_uris = []
    if pgsql.master is None:
        raise MainDatabaseNotAvailable("main database not ready (pgsql.master is None)")
    return pgsql.master.uri, slave_uris
