# Copyright 2019-2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import subprocess

import distro

from charmhelpers.core import hookenv, host, templating

import ols.base


def render(source=None, target=None, context=None):
    """Render a template to a path, returning whether the contents changed.

    Takes some care to avoid writing the file is nothing would change. This
    helps some utilities from needlessly restarting if the timestamp has
    changed but the content has not.
    """
    temp = target + '.tmp'
    changed = False
    try:
        templating.render(source, temp, context)
        old = host.file_hash(target) if os.path.exists(target) else None
        new = host.file_hash(temp)

        if old != new:
            os.rename(temp, target)
            changed = True
    finally:
        if os.path.exists(temp):
            os.unlink(temp)

    return changed


def render_logrotate(target, context, template='logrotate.tmpl'):
    # use old style rsyslog rotate if older than 22.04
    local_context = context.copy()
    local_context['old_rsyslog_rotate'] = int(distro.major_version()) < 22
    render(source=template, target=target, context=local_context)


def setup_charm_service(exec_path, exec_args=None):
    name = ols.base.service_name()
    return setup_systemd_unit(
        name=name,
        user=ols.base.user(),
        paths=ols.base.service_paths(),
        exec_path=exec_path,
        exec_args=exec_args,
        env=ols.base.env_vars(),
    )


def setup_systemd_unit(
    name,
    user,
    paths,
    exec_path,
    exec_args=None,
    env=None,
    terminate='TERM',
    reload='HUP',
):
    """Set up systemd unit."""

    exec_full = exec_path
    if exec_args is not None:
        exec_full += ' ' + exec_args

    ctx = dict(
        name=name,
        user=user,
        group=user,
        exec_path=exec_path,
        exec_full=exec_full,
        exec_name=os.path.basename(exec_path),
        log_path=paths['log'],
        working_dir=paths['working_dir'],
        terminate=terminate,
        reload=reload,
        environ=env or {},
        env_path=paths['env'],
    )

    env_changed = render('env.tmpl', paths['env'], ctx)
    render('logrotate.tmpl', paths['logrotate'], ctx)
    systemd_changed = render('systemd.tmpl', paths['systemd'], ctx)
    syslog_changed = render('syslog.tmpl', paths['syslog'], ctx)

    # systemd decided to revert group write perms on /var/log, so...
    if not os.path.exists(paths['log']):
        open(paths['log'], 'wb').close()
    host.chownr(paths['log'], 'syslog', user, chowntopdir=True)
    os.chmod(paths['log'], 0o644)

    if syslog_changed:
        subprocess.check_call(['systemctl', 'force-reload', 'rsyslog'])

    if systemd_changed:
        # reload systemd to get new configs applied
        hookenv.log('reloading systemd as {} changed'.format(paths['systemd']))
        subprocess.check_call(['systemctl', 'daemon-reload'])
        subprocess.check_call(['systemctl', 'enable', name])

    return env_changed or systemd_changed
