# Copyright 2022 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from charmhelpers.core import hookenv
from charms.launchpad.base import (
    change_shell,
    configure_rsync,
    ensure_lp_directories,
    get_service_config,
)
from charms.launchpad.payload import configure_lazr
from charms.reactive import (
    clear_flag,
    endpoint_from_flag,
    hook,
    set_flag,
    when,
    when_any,
    when_none,
    when_not,
    when_not_all,
)
from ols import base


@when("rabbitmq.connected")
def prepare_rabbitmq():
    rabbitmq = endpoint_from_flag("rabbitmq.connected")
    config = hookenv.config()
    rabbitmq.request_access(config["rabbitmq_user"], config["domain"])


def get_rabbitmq_uris(rabbitmq):
    config = hookenv.config()
    if config["rabbitmq_broker_urls"]:
        yield from config["rabbitmq_broker_urls"].split()
    else:
        for conversation in rabbitmq.conversations():
            for relation_id in conversation.relation_ids:
                for unit in hookenv.related_units(relation_id):
                    hostname = hookenv.relation_get(
                        "private-address", unit, relation_id
                    )
                    vhost = rabbitmq.vhost()
                    username = rabbitmq.username()
                    password = rabbitmq.password()
                    yield f"amqp://{username}:{password}@{hostname}/{vhost}"


@when_any("rabbitmq.available", "config.set.rabbitmq_broker_urls")
def rabbitmq_available():
    set_flag("launchpad.rabbitmq.available")


@when_none("rabbitmq.available", "config.set.rabbitmq_broker_urls")
def rabbitmq_unavailable():
    clear_flag("launchpad.rabbitmq.available")


@when("launchpad.payload.configured", "launchpad.rabbitmq.available")
@when_not("launchpad.base.configured")
def configure():
    rabbitmq = endpoint_from_flag("rabbitmq.available")
    # Interactive use shouldn't be frequent, but it's still needed
    # sometimes, so make it less annoying.
    change_shell(base.user(), "/bin/bash")
    ensure_lp_directories()
    config = get_service_config()
    config["rabbitmq_broker_urls"] = sorted(get_rabbitmq_uris(rabbitmq))
    configure_lazr(
        config,
        "launchpad-base-lazr.conf",
        "launchpad-base-lazr.conf",
    )
    configure_lazr(
        config,
        "launchpad-base-secrets-lazr.conf",
        "launchpad-base-secrets-lazr.conf",
        secret=True,
    )
    configure_rsync(
        config, "launchpad-base-rsync.conf", "010-launchpad-base.conf"
    )
    set_flag("launchpad.base.configured")


@when("launchpad.base.configured")
@when_not_all("launchpad.payload.configured", "launchpad.rabbitmq.available")
def deconfigure():
    clear_flag("launchpad.base.configured")


@hook("upgrade-charm")
def upgrade_charm():
    clear_flag("launchpad.base.configured")


@when("config.changed")
def config_changed():
    clear_flag("launchpad.base.configured")


@hook("{requires:rabbitmq}-relation-changed")
def rabbitmq_relation_changed(*args):
    clear_flag("launchpad.base.configured")
