# Copyright 2019-2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import os
import subprocess
import tarfile

import distro
import reactive.ols  # noqa


from collections import OrderedDict
from unittest.mock import patch, call, Mock

from charms import layer
from charms.reactive import bus, set_flag
from charmhelpers.core import hookenv, unitdata, host

from fixtures import EnvironmentVariable, MockPatchObject, TempDir
from systemfixtures import FakeProcesses

from ols import base, render
from base_fixtures import CharmEnvFixture, ReactiveCharmEnvFixture, BaseTestCase


CHARM_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '../'))


class BaseTest(BaseTestCase):
    swift_options = {
        'swift_username': 'some-username',
        'swift_password': 'some-password',
        'swift_auth_url': 'http://keystone.example',
        'swift_region_name': 'some-region',
        'swift_tenant_name': 'some-tenant',
        'swift_container_name': 'some-container',
    }

    def setUp(self):
        super().setUp()
        self.charm_dir = '/charm'
        self.fakes = self.useFixture(CharmEnvFixture(self.charm_dir))

    def test_layer_options(self):
        self.assertEqual(base._layer_options(), layer.options().get('ols', {}))

    def test_service_name(self):
        self.assertEqual(base.service_name(), 'testservice')

    def test_user(self):
        self.assertEqual(base.user(), 'ols')

    def test_tarball_payload(self):
        self.assertFalse(base.tarball_payload())
        self.fakes.layer_config.update({'ols': {'tarball_payload': True}})
        self.assertTrue(base.tarball_payload())

    def test_symlink_switch_payload(self):
        self.assertFalse(base.symlink_switch_payload())
        self.fakes.layer_config.update({'ols': {'symlink_switch_payload': True}})
        self.assertTrue(base.symlink_switch_payload())

    def test_python_bin(self):
        self.assertEqual(base.python_bin(), 'python3')

    def test_use_system_site_packages(self):
        self.assertFalse(base.use_system_site_packages())

    def test_root_dir(self):
        self.assertEqual(base.root_dir(), '/srv')

    def test_base_dir(self):
        self.assertEqual(base.base_dir(), '/srv/testservice')

    def test_etc_dir(self):
        self.assertEqual(base.etc_dir(), '/srv/testservice/etc')

    def test_payloads_dir(self):
        self.assertEqual(base.payloads_dir(), '/srv/testservice/payloads')

    def test_code_dir(self):
        self.assertEqual(base.code_dir(), '/srv/testservice/code')

    def test_logs_dir(self):
        self.assertEqual(base.logs_dir(), '/srv/testservice/logs')

    def test_readonly_dirs(self):
        self.assertEqual(
            base.readonly_dirs(),
            [
                '/srv',
                '/srv/testservice',
                '/srv/testservice/etc',
                '/srv/testservice/payloads',
            ],
        )

    def test_writable_dirs(self):
        self.assertEqual(base.writable_dirs(), ['/srv/testservice/logs'])

    def test_service_config_path(self):
        self.assertEqual(
            base.service_config_path(), '/srv/testservice/etc/service.conf'
        )

    def test_service_config_filename(self):
        self.assertEqual(base.service_config_filename(), 'service.conf')

    def test_config_files(self):
        self.assertEqual(base.config_files(), [base.service_config_path()])

    def test_service_paths(self):
        name = base.service_name()
        expected = {
            'config': base.service_config_path(),
            'working_dir': base.code_dir(),
            'log': '/var/log/{}.log'.format(name),
            'env': base.etc_dir() + '/{}.env'.format(name),
            'systemd': '/lib/systemd/system/{}.service'.format(name),
            'syslog': '/etc/rsyslog.d/22-{}.conf'.format(name),
            'logrotate': '/etc/logrotate.d/{}'.format(name),
        }
        self.assertDictEqual(base.service_paths(), expected)

    def test_payload_name(self):
        self.assertEqual(base.payload_name(), 'testservice')

    def test_payload_dir(self):
        self.assertEqual(base.payload_dir(), os.path.join(self.charm_dir, 'files'))

    def test_version_info_path(self):
        self.assertEqual(
            base.version_info_path(), '/srv/testservice/code/version-info.txt'
        )

    def test_env_vars(self):
        # override env vars and config with empty defaults
        new_env = {
            'JUJU_MACHINE_ID': 'machine-0',
            'JUJU_UNIT_NAME': 'stubcharm/0',
            'JUJU_MODEL_NAME': 'mymodel',
        }
        for k, v in new_env.items():
            self.useFixture(EnvironmentVariable(k, v))

        new_config = {
            'sentry_dsn': 'sentry://something',
            'statsd_dsn': 'statsd://something',
            'debuglog': 'path/to/logfile',
        }
        self.fakes.config.update(new_config)
        expected = OrderedDict(
            [
                ('SERVICE_CONFIG', '/srv/testservice/etc/service.conf'),
                ('SERVICE_ENVIRONMENT', 'production'),
                ('SERVICE_DIR', '/srv/testservice/code'),
                ('JUJU_MODEL', 'mymodel'),
                ('JUJU_UNIT', 'stubcharm/0'),
                ('JUJU_MACHINE', 'machine-0'),
                ('STATSD_DSN', 'statsd://something'),
                ('SENTRY_DSN', 'sentry://something'),
                ('TALISKER_NETWORKS', ''),
                ('TALISKER_SOFT_REQUEST_TIMEOUT', '-1'),
                ('TALISKER_SLOWQUERY_THRESHOLD', '-1'),
                ('DEBUGLOG', 'path/to/logfile'),
            ]
        )
        self.assertDictEqual(base.env_vars(), expected)

    def test_env_vars_trusted_networks_ordering(self):
        self.maxDiff = None
        # override env vars and config with empty defaults
        new_env = {
            'JUJU_MACHINE_ID': 'machine-0',
            'JUJU_UNIT_NAME': 'stubcharm/0',
            'JUJU_MODEL_NAME': 'mymodel',
        }
        for k, v in new_env.items():
            self.useFixture(EnvironmentVariable(k, v))

        new_config = {
            'sentry_dsn': 'sentry://something',
            'statsd_dsn': 'statsd://something',
            'debuglog': 'path/to/logfile',
            'trusted_networks': '10.131.37.188 10.131.37.0/24',
        }
        self.fakes.config.update(new_config)
        expected = OrderedDict(
            [
                ('SERVICE_CONFIG', '/srv/testservice/etc/service.conf'),
                ('SERVICE_ENVIRONMENT', 'production'),
                ('SERVICE_DIR', '/srv/testservice/code'),
                ('JUJU_MODEL', 'mymodel'),
                ('JUJU_UNIT', 'stubcharm/0'),
                ('JUJU_MACHINE', 'machine-0'),
                ('STATSD_DSN', 'statsd://something'),
                ('SENTRY_DSN', 'sentry://something'),
                ('TALISKER_NETWORKS', '10.131.37.0/24 10.131.37.188'),
                ('TALISKER_SOFT_REQUEST_TIMEOUT', '-1'),
                ('TALISKER_SLOWQUERY_THRESHOLD', '-1'),
                ('DEBUGLOG', 'path/to/logfile'),
            ]
        )
        self.assertDictEqual(base.env_vars(), expected)

    @patch.object(hookenv, 'log')
    def test_set_extra_trusted_networks(self, log):
        unit_data = unitdata.kv()
        expected = ['192.168.1.21', '10.5.25.0/16']
        base.set_extra_trusted_networks('foo', expected)
        log.assert_called_once_with(
            "Setting extra trusted networks: foo=['192.168.1.21', '10.5.25.0/16']"
        )
        value = unit_data.get('extra_trusted_networks.foo')
        self.assertEqual(value, expected)

    @patch.object(hookenv, 'log')
    def test_reset_extra_trusted_networks(self, log):
        unit_data = unitdata.kv()
        expected = ['192.168.1.21', '10.5.25.0/16']
        base.set_extra_trusted_networks('foo', expected)
        log.assert_called_with(
            "Setting extra trusted networks: foo=['192.168.1.21', '10.5.25.0/16']"
        )
        value = unit_data.get('extra_trusted_networks.foo')
        self.assertEqual(value, expected)
        base.reset_extra_trusted_networks('foo')
        value = unit_data.get('extra_trusted_networks.foo')
        self.assertEqual(value, None)
        log.assert_called_with("Removing extra trusted networks: foo")

    @patch.object(host, 'adduser')
    @patch.object(host, 'add_group')
    @patch.object(hookenv, 'log')
    @patch.object(host, 'mkdir')
    def test_ensure_directories(self, mkdir, log, add_group, adduser):
        base.ensure_directories(base.readonly_dirs(), base.writable_dirs(), base.user())
        add_group.assert_called_once_with(base.user())
        adduser.assert_called_once_with(base.user(), primary_group=base.user())
        mkdir_calls = []
        for directory in base.readonly_dirs():
            mkdir_calls.append(call(directory, group=base.user(), perms=0o755))
        for directory in base.writable_dirs():
            mkdir_calls.append(call(directory, group=base.user(), perms=0o775))
        mkdir.assert_has_calls(mkdir_calls)

    def test_fetch_payload_no_tarball_payload(self):
        with self.assertRaises(base.PayloadError) as pe:
            base.fetch_payload()
        self.assertEqual(str(pe.exception), 'Can only fetch tarball payloads')

    def test_fetch_payload_no_build_label(self):
        self.fakes.layer_config.update({'ols': {'tarball_payload': True}})
        with self.assertRaises(base.PayloadError) as pe:
            base.fetch_payload()
        self.assertEqual(
            str(pe.exception), 'build_label unset, so cannot fetch payload'
        )

    def test_fetch_payload_anonymous_swift(self):
        self.fakes.layer_config.update({'ols': {'tarball_payload': True}})
        self.fakes.config.update(
            {
                'build_label': '123',
                'swift_container_name': 'some-container',
                'swift_storage_url': 'http://swift.example',
            }
        )
        ps = self.useFixture(FakeProcesses())
        wget_ok = Mock()
        wget_ok.return_value = {'returncode': 0}
        ps.add(wget_ok, name='wget')
        expected_archive_path = '/srv/testservice/payloads/123.tar.gz'
        self.assertEqual(expected_archive_path, base.fetch_payload())
        wget_ok.assert_called_once_with(
            {
                'args': [
                    'wget',
                    '-O',
                    expected_archive_path,
                    'http://swift.example/some-container/'
                    'testservice-builds/123/testservice.tar.gz',
                ]
            }
        )

    def test_fetch_payload_resource(self):
        self.fakes.layer_config.update({'ols': {'tarball_payload': True}})
        self.fakes.config.update({'build_label': '123'})
        os.makedirs(base.payloads_dir())
        temp_dir = self.useFixture(TempDir()).path
        resource_path = os.path.join(temp_dir, 'testservice.tar.gz')
        with open(resource_path, 'wb') as f:
            f.write(b'payload')
        ps = self.useFixture(FakeProcesses())
        resource_get_ok = Mock(
            return_value={'stdout': io.BytesIO(resource_path.encode('UTF-8'))}
        )
        ps.add(resource_get_ok, name='resource-get')
        expected_archive_path = '/srv/testservice/payloads/123.tar.gz'
        self.assertEqual(expected_archive_path, base.fetch_payload())
        resource_get_ok.assert_called_once_with(
            {'args': ['resource-get', 'testservice'], 'stdout': subprocess.PIPE}
        )

    def test_fetch_payload_no_resource(self):
        self.fakes.layer_config.update({'ols': {'tarball_payload': True}})
        self.fakes.config.update({'build_label': '123'})
        ps = self.useFixture(FakeProcesses())
        ps.add(Mock(return_value={'returncode': 1}), name='resource-get')
        with self.assertRaises(base.PayloadError) as pe:
            base.fetch_payload()
        self.assertEqual(str(pe.exception), 'No payload available')

    def test_unlink_force_file_present(self):
        temp_dir = self.useFixture(TempDir()).path
        path = os.path.join(temp_dir, "file")
        with open(path, "w"):
            pass
        base.unlink_force(path)
        self.assertFalse(os.path.exists(path))

    def test_unlink_force_file_missing(self):
        temp_dir = self.useFixture(TempDir()).path
        path = os.path.join(temp_dir, "file")
        base.unlink_force(path)
        self.assertFalse(os.path.exists(path))

    def test_unlink_force_oserror(self):
        temp_dir = self.useFixture(TempDir()).path
        path = os.path.join(temp_dir, "dir")
        os.mkdir(path)
        self.assertRaises(OSError, base.unlink_force, path)

    def test_symlink_force_file_present(self):
        temp_dir = self.useFixture(TempDir()).path
        path = os.path.join(temp_dir, "link")
        with open(path, "w"):
            pass
        base.symlink_force("source", path)
        self.assertTrue(os.path.islink(path))
        self.assertEqual("source", os.readlink(path))

    def test_symlink_force_link_present(self):
        temp_dir = self.useFixture(TempDir()).path
        path = os.path.join(temp_dir, "link")
        os.symlink("old", path)
        base.symlink_force("source", path)
        self.assertTrue(os.path.islink(path))
        self.assertEqual("source", os.readlink(path))

    def test_symlink_force_missing(self):
        temp_dir = self.useFixture(TempDir()).path
        path = os.path.join(temp_dir, "link")
        base.symlink_force("source", path)
        self.assertTrue(os.path.islink(path))
        self.assertEqual("source", os.readlink(path))

    @patch.object(hookenv, 'log')
    def test_prune_payloads(self, log):
        payloads_dir = base.payloads_dir()
        os.makedirs(payloads_dir)
        for label in ('99', '100', '101'):
            with open(os.path.join(payloads_dir, "{}.tar.gz".format(label)), "w"):
                pass
            os.mkdir(os.path.join(payloads_dir, label))
        base.prune_payloads(keep_labels=['100', '101'])
        self.assertEqual(
            sorted(['100', '100.tar.gz', '101', '101.tar.gz']),
            sorted(os.listdir(payloads_dir)),
        )
        log.assert_has_calls(
            [
                call(
                    'Purging old payload tarball in {}/99.tar.gz'.format(payloads_dir)
                ),
                call('Purging old build in {}/99'.format(payloads_dir)),
            ],
            any_order=True,
        )

    @patch.object(host, 'chownr')
    @patch.object(base, 'payload_dir')
    @patch.object(hookenv, 'log')
    def test_install_payload(self, log, payload_dir, chownr):
        # override payload_dir with real CHARM_DIR/files  as using overlayed
        # src and dst in shutil.copytree fails. To reproduce change CHARM_DIR
        # with self.charm_dir
        # https://github.com/testing-cabal/systemfixtures/issues/7
        payload_dir.return_value = os.path.join(CHARM_DIR, 'files')
        base.install_payload(base.payload_name(), base.code_dir(), 'root')
        self.assertTrue(os.path.exists(base.code_dir()))
        files = os.path.join(self.charm_dir, 'files', 'testservice')
        self.assertCountEqual(os.listdir(files), os.listdir(base.code_dir()))
        log.assert_called_with('Copying charm payload to {}'.format(base.code_dir()))
        chownr.assert_called_with(base.code_dir(), 'root', 'root', chowntopdir=True)
        # call it again, should be a noop
        base.install_payload(base.payload_name(), base.code_dir(), 'root')
        log.assert_called_with("Payload already extracted in a previous run")

    @patch.object(host, 'chownr')
    @patch.object(base, 'base_dir')
    @patch.object(hookenv, 'log')
    def test_install_payload_symlink_switch(self, log, base_dir, chownr):
        # systemfixtures doesn't mock tarfile.bltn_open, so use a real
        # temporary directory for base_dir.
        base_dir.return_value = self.useFixture(TempDir()).path
        self.fakes.layer_config.update(
            {'ols': {'tarball_payload': True, 'symlink_switch_payload': True}}
        )
        self.fakes.config.update(
            {
                'build_label': '123',
                'swift_container_name': 'some-container',
                'swift_storage_url': 'http://swift.example',
            }
        )
        os.makedirs(base.payloads_dir())
        ps = self.useFixture(FakeProcesses())

        def wget_side_effect(proc_args):
            with tarfile.open(proc_args['args'][2], 'w') as tar:
                build_label = hookenv.config()['build_label'].encode('UTF-8')
                tarinfo = tarfile.TarInfo('./build-label')
                tarinfo.mode = 0o644
                tarinfo.size = len(build_label)
                tar.addfile(tarinfo, io.BytesIO(build_label))
            return {'returncode': 0}

        wget = Mock(side_effect=wget_side_effect)
        ps.add(wget, name='wget')
        base.install_payload(base.payload_name(), base.code_dir(), 'root')
        self.assertEqual(['123', '123.tar.gz'], sorted(os.listdir(base.payloads_dir())))
        this_payload_dir = os.path.join(base.payloads_dir(), '123')
        self.assertTrue(os.path.islink(base.code_dir()))
        self.assertEqual(this_payload_dir, os.path.realpath(base.code_dir()))
        with open(os.path.join(base.code_dir(), 'build-label')) as f:
            self.assertEqual('123', f.read())
        log.assert_called_with(
            'Extracting charm payload to {}'.format(this_payload_dir)
        )
        chownr.assert_called_with(this_payload_dir, 'root', 'root', chowntopdir=True)
        # call it again, should be a noop
        base.install_payload(base.payload_name(), base.code_dir(), 'root')
        log.assert_called_with("Payload already extracted in a previous run")
        self.assertEqual(['123', '123.tar.gz'], sorted(os.listdir(base.payloads_dir())))
        # and again with a different build label
        self.fakes.config.update({'build_label': '124'})
        base.install_payload(base.payload_name(), base.code_dir(), 'root')
        self.assertEqual(
            ['123', '123.tar.gz', '124', '124.tar.gz'],
            sorted(os.listdir(base.payloads_dir())),
        )
        this_payload_dir = os.path.join(base.payloads_dir(), '124')
        self.assertTrue(os.path.islink(base.code_dir()))
        self.assertEqual(this_payload_dir, os.path.realpath(base.code_dir()))
        with open(os.path.join(base.code_dir(), 'build-label')) as f:
            self.assertEqual('124', f.read())
        log.assert_called_with(
            'Extracting charm payload to {}'.format(this_payload_dir)
        )
        chownr.assert_called_with(this_payload_dir, 'root', 'root', chowntopdir=True)
        # and again, should be a noop
        base.install_payload(base.payload_name(), base.code_dir(), 'root')
        log.assert_called_with("Payload already extracted in a previous run")
        self.assertEqual(
            ['123', '123.tar.gz', '124', '124.tar.gz'],
            sorted(os.listdir(base.payloads_dir())),
        )
        # and one last time, should prune 123
        self.fakes.config.update({'build_label': '125'})
        base.install_payload(base.payload_name(), base.code_dir(), 'root')
        self.assertEqual(
            ['124', '124.tar.gz', '125', '125.tar.gz'],
            sorted(os.listdir(base.payloads_dir())),
        )
        this_payload_dir = os.path.join(base.payloads_dir(), '125')
        self.assertTrue(os.path.islink(base.code_dir()))
        self.assertEqual(this_payload_dir, os.path.realpath(base.code_dir()))
        with open(os.path.join(base.code_dir(), 'build-label')) as f:
            self.assertEqual('125', f.read())
        log.assert_has_calls(
            [
                call('Extracting charm payload to {}'.format(this_payload_dir)),
                call(
                    'Purging old payload tarball in {}/123.tar.gz'.format(
                        base.payloads_dir()
                    )
                ),
                call('Purging old build in {}/123'.format(base.payloads_dir())),
            ],
            any_order=True,
        )
        chownr.assert_called_with(this_payload_dir, 'root', 'root', chowntopdir=True)

    @patch.object(hookenv, 'log')
    def test_create_virtualenv(self, log):
        env_dir = os.path.join(base.code_dir(), 'env')
        wheels_dir = os.path.join(base.code_dir(), 'wheels')
        requirements = os.path.join(base.code_dir(), 'requirements.txt')
        ps = FakeProcesses()
        self.useFixture(ps)
        apt_get_ok = Mock()
        apt_get_ok.return_value = {'returncode': 0}
        ps.add(apt_get_ok, name='apt-get')
        venv_ok = Mock()
        venv_ok.return_value = {'returncode': 0}
        if int(distro.major_version()) >= 22:
            ps.add(venv_ok, name='python3')
        else:
            ps.add(venv_ok, name='virtualenv')
        pip = Mock()
        pip.return_value = {'returncode': 0}
        pip_bin = '%s/bin/pip' % (env_dir)
        ps.add(pip, name=pip_bin)

        base.create_virtualenv(wheels_dir, base.code_dir(), base.python_bin())

        if int(distro.major_version()) >= 22:
            venv_ok.assert_called_once_with(
                {'args': ['python3', '-m', 'venv', '/srv/testservice/code/env']}
            )
            install_msg = 'Use python3 venv for virtualenv creation'
        else:
            venv_ok.assert_called_once_with(
                {
                    'args': [
                        'virtualenv',
                        '--no-download',
                        '--extra-search-dir=/srv/testservice/code/wheels',
                        '-p',
                        'python3',
                        '--extra-search-dir=/srv/testservice/code/wheels',
                        '/srv/testservice/code/env',
                    ]
                }
            )
            install_msg = 'Installing python packages into the venv'

        pip.assert_called_once_with(
            {
                'args': [
                    pip_bin,
                    'install',
                    '--no-index',
                    '--find-links={}'.format(wheels_dir),
                    '-r',
                    requirements,
                ]
            }
        )
        log.assert_has_calls(
            [call(install_msg), call('Installing from local wheelhouse.')]
        )

    @patch.object(hookenv, 'log')
    def test_create_virtualenv_with_setup_py(self, log):
        env_dir = os.path.join(base.code_dir(), 'env')
        wheels_dir = os.path.join(base.code_dir(), 'wheels')
        requirements = os.path.join(base.code_dir(), 'requirements.txt')
        setup_py = os.path.join(base.code_dir(), 'setup.py')
        os.makedirs(base.code_dir())
        with open(setup_py, 'w'):
            pass
        ps = FakeProcesses()
        self.useFixture(ps)
        apt_get_ok = Mock()
        apt_get_ok.return_value = {'returncode': 0}
        ps.add(apt_get_ok, name='apt-get')
        venv_ok = Mock()
        venv_ok.return_value = {'returncode': 0}
        if int(distro.major_version()) >= 22:
            ps.add(venv_ok, name='python3')
        else:
            ps.add(venv_ok, name='virtualenv')
        pip = Mock()
        pip.return_value = {'returncode': 0}
        pip_bin = '%s/bin/pip' % (env_dir)
        ps.add(pip, name=pip_bin)

        base.create_virtualenv(wheels_dir, base.code_dir(), base.python_bin())

        if int(distro.major_version()) >= 22:
            venv_ok.assert_called_once_with(
                {'args': ['python3', '-m', 'venv', '/srv/testservice/code/env']}
            )
            install_msg = 'Use python3 venv for virtualenv creation'
        else:
            venv_ok.assert_called_once_with(
                {
                    'args': [
                        'virtualenv',
                        '--no-download',
                        '--extra-search-dir=/srv/testservice/code/wheels',
                        '-p',
                        'python3',
                        '--extra-search-dir=/srv/testservice/code/wheels',
                        '/srv/testservice/code/env',
                    ]
                }
            )
            install_msg = 'Installing python packages into the venv'

        pip.assert_called_once_with(
            {
                'args': [
                    pip_bin,
                    'install',
                    '--no-index',
                    '--find-links={}'.format(wheels_dir),
                    '-r',
                    requirements,
                    '-e',
                    base.code_dir(),
                ]
            }
        )
        log.assert_has_calls(
            [call(install_msg), call('Installing from local wheelhouse.')]
        )

    def test_render_template(self):
        tmpl = """
        {% if True %}
        {{foo}}
            {% if True %}
        {{foo}}
            {% endif %}
        {% endif %}
        """
        expected = """
        bar
        bar
        """
        actual = base.render_template(tmpl, {'foo': 'bar'})
        self.assertEqual(expected, actual)

    @patch.object(hookenv, 'open_port')
    @patch.object(hookenv, 'close_port')
    def test_check_port(self, close_port, open_port):
        unit_data = unitdata.kv()
        base.check_port('test', 8080)
        open_port.assert_called_with(8080)
        self.assertEqual(unit_data.get('test.port'), 8080)
        base.check_port('test', 8081)
        close_port.assert_called_once_with(8080)
        open_port.assert_called_with(8081)
        self.assertEqual(unit_data.get('test.port'), 8081)


class SwiftPayloadTest(BaseTestCase):
    swift_options = {
        'swift_username': 'some-username',
        'swift_password': 'some-password',
        'swift_auth_url': 'http://keystone.example',
        'swift_region_name': 'some-region',
        'swift_tenant_name': 'some-tenant',
        'swift_container_name': 'some-container',
    }

    def setUp(self):
        super().setUp()
        self.charm_dir = '/charm'
        self.fakes = self.useFixture(CharmEnvFixture(self.charm_dir))
        ps = self.useFixture(FakeProcesses())
        self.apt_get_ok = Mock()
        self.apt_get_ok.return_value = {'returncode': 0}
        ps.add(self.apt_get_ok, name='apt-get')
        self.swift_ok = Mock()
        self.swift_ok.return_value = {'returncode': 0}
        ps.add(self.swift_ok, name='swift')
        self.expected_archive_path = '/srv/testservice/payloads/123.tar.gz'

    def update_config(self, layer_config, config, swift_options):
        self.fakes.layer_config.update(layer_config)
        self.fakes.config.update(swift_options)
        self.fakes.config.update(config)

    def assert_apt_get_called(self):
        self.assertEqual(
            self.apt_get_ok.call_args_list[0].args[0]['args'],
            [
                'apt-get',
                '--option=Dpkg::Options::=--force-confold',
                '--assume-yes',
                'install',
                'python3-swiftclient',
            ],
        )

    def test_fetch_payload_authenticated_swift(self):
        self.update_config(
            {'ols': {'tarball_payload': True}},
            self.swift_options,
            {'build_label': '123'},
        )
        self.assertEqual(self.expected_archive_path, base.fetch_payload())

        self.assert_apt_get_called()

        self.swift_ok.assert_called_once_with(
            {
                'args': [
                    'swift',
                    '--os-username=some-username',
                    '--os-password=some-password',
                    '--os-auth-url=http://keystone.example',
                    '--os-region-name=some-region',
                    '--os-tenant-name=some-tenant',
                    'download',
                    '--output={}'.format(self.expected_archive_path),
                    'some-container',
                    'testservice-builds/123/testservice.tar.gz',
                ]
            }
        )

    def test_fetch_payload_authenticated_swift_auth_v3(self):
        self.update_config(
            {'ols': {'tarball_payload': True}},
            self.swift_options,
            {
                'build_label': '123',
                'swift_auth_version': '3',
                'swift_user_domain_name': 'not-default',
                'swift_project_domain_name': 'not-default',
            },
        )
        self.assertEqual(self.expected_archive_path, base.fetch_payload())
        self.assert_apt_get_called()
        self.swift_ok.assert_called_once_with(
            {
                'args': [
                    'swift',
                    '--os-username=some-username',
                    '--os-password=some-password',
                    '--os-auth-url=http://keystone.example',
                    '--os-region-name=some-region',
                    '--os-tenant-name=some-tenant',
                    '--auth-version=3',
                    '--os-user-domain-name=not-default',
                    '--os-project-domain-name=not-default',
                    'download',
                    '--output={}'.format(self.expected_archive_path),
                    'some-container',
                    'testservice-builds/123/testservice.tar.gz',
                ]
            }
        )

    def test_fetch_payload_authenticated_swift_tarball_name(self):
        self.update_config(
            {
                'ols': {
                    'tarball_payload': True,
                    'tarball_payload_name': 'testservice',
                    'service_name': 'testservice-foo',
                }
            },
            self.swift_options,
            {
                'build_label': '123',
                'swift_auth_version': '3',
                'swift_user_domain_name': 'not-default',
                'swift_project_domain_name': 'not-default',
            },
        )
        self.expected_archive_path = '/srv/testservice-foo/payloads/123.tar.gz'
        self.assertEqual(self.expected_archive_path, base.fetch_payload())
        self.swift_ok.assert_called_once_with(
            {
                'args': [
                    'swift',
                    '--os-username=some-username',
                    '--os-password=some-password',
                    '--os-auth-url=http://keystone.example',
                    '--os-region-name=some-region',
                    '--os-tenant-name=some-tenant',
                    '--auth-version=3',
                    '--os-user-domain-name=not-default',
                    '--os-project-domain-name=not-default',
                    'download',
                    '--output={}'.format(self.expected_archive_path),
                    'some-container',
                    'testservice-builds/123/testservice.tar.gz',
                ]
            }
        )

    def test_fetch_payload_authenticated_swift_storage_url(self):
        self.update_config(
            {'ols': {'tarball_payload': True}},
            self.swift_options,
            {
                'build_label': '123',
                'swift_auth_version': '3',
                'swift_user_domain_name': 'not-default',
                'swift_project_domain_name': 'not-default',
                'swift_storage_url': 'http://foo.bar/v1/AUTH_foo',
            },
        )
        self.assertEqual(self.expected_archive_path, base.fetch_payload())
        self.assert_apt_get_called()
        self.swift_ok.assert_called_once_with(
            {
                'args': [
                    'swift',
                    '--os-username=some-username',
                    '--os-password=some-password',
                    '--os-auth-url=http://keystone.example',
                    '--os-region-name=some-region',
                    '--os-tenant-name=some-tenant',
                    '--auth-version=3',
                    '--os-user-domain-name=not-default',
                    '--os-project-domain-name=not-default',
                    '--os-storage-url=http://foo.bar/v1/AUTH_foo',
                    'download',
                    '--output={}'.format(self.expected_archive_path),
                    'some-container',
                    'testservice-builds/123/testservice.tar.gz',
                ]
            }
        )


class ReactiveTest(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.charm_dir = '/charm'
        self.fakes = self.useFixture(ReactiveCharmEnvFixture(charm_dir=self.charm_dir))
        chownr = self.useFixture(MockPatchObject(host, 'chownr'))
        self.chownr = chownr.mock

    @patch.object(hookenv, 'application_version_set')
    def test_install_service(self, app_ver_set):
        bus.dispatch()
        self.chownr.assert_called_with(
            base.code_dir(), 'root', 'root', chowntopdir=True
        )
        self.assertEqual(
            ['ols.configured', 'ols.service.installed'],
            sorted(list(bus.get_states().keys())),
        )
        self.fakes.status_set.assert_has_calls(
            [
                call('waiting', 'Service installed, but not configured.'),
                call('waiting', 'ols layer configured'),
            ]
        )
        app_ver_set.assert_not_called()

    def test_install_service_queued_install(self):
        set_flag('apt.queued_installs')
        bus.dispatch()
        self.assertNotIn('ols.service.installed', bus.get_states().keys())

    @patch.object(hookenv, 'application_version_set')
    def test_handle_config_changed(self, app_ver_set):
        set_flag('ols.configured')
        set_flag('service.configured')
        set_flag('ols.service.installed')
        bus.dispatch()
        self.assertEqual(
            ['ols.configured', 'ols.service.installed', 'service.configured'],
            sorted(list(bus.get_states().keys())),
        )
        set_flag('config.changed')
        bus.dispatch()
        self.assertNotIn('service.configured', bus.get_states().keys())
        self.chownr.assert_not_called()
        self.fakes.status_set.assert_called_once_with('waiting', 'ols layer configured')

    def test_reconfigure(self):
        set_flag('service.configured')
        set_flag('ols.service.installed')
        bus.dispatch()
        self.assertNotIn('service.configured', bus.get_states().keys())
        self.chownr.assert_not_called()
        self.fakes.status_set.assert_called_once_with('waiting', 'ols layer configured')

    def test_ols_configured(self):
        set_flag('ols.service.installed')
        bus.dispatch()
        self.assertEqual(
            ['ols.configured', 'ols.service.installed'],
            sorted(list(bus.get_states().keys())),
        )
        self.chownr.assert_not_called()
        self.fakes.status_set.assert_called_once_with('waiting', 'ols layer configured')

    @patch.object(reactive.ols, 'ensure_directories')
    @patch.object(reactive.ols, 'install_payload')
    @patch.object(hookenv, 'application_version_set')
    def test_handle_build_label_change(
        self, app_ver_set, install_payload, ensure_directories
    ):
        set_flag('ols.configured')
        set_flag('service.configured')
        set_flag('ols.service.installed')
        bus.dispatch()
        self.assertEqual(
            ['ols.configured', 'ols.service.installed', 'service.configured'],
            sorted(list(bus.get_states().keys())),
        )
        set_flag('config.changed.build_label')
        bus.dispatch()
        self.assertNotIn('service.configured', bus.get_states().keys())
        # check install service related calls are there
        ensure_directories.assert_called_once_with(
            base.readonly_dirs(), base.writable_dirs(), base.user()
        )
        install_payload.assert_called_once_with(
            base.payload_name(), base.code_dir(), 'root'
        )
        self.fakes.status_set.assert_has_calls(
            [
                call('waiting', 'Service installed, but not configured.'),
                call('waiting', 'ols layer configured'),
            ]
        )
        app_ver_set.assert_not_called()

    @patch.object(reactive.ols, 'ensure_directories')
    @patch.object(reactive.ols, 'install_payload')
    @patch.object(hookenv, 'application_version_set')
    def test_handle_build_label_not_set_symlink_switch(
        self, app_ver_set, install_payload, ensure_directories
    ):
        self.fakes.charm_env.layer_config.update(
            {'ols': {'tarball_payload': True, 'symlink_switch_payload': True}}
        )
        bus.dispatch()
        self.fakes.status_set.assert_has_calls(
            [call('waiting', 'build_label not specified')]
        )
        install_payload.assert_not_called()
        app_ver_set.assert_not_called()

    @patch.object(reactive.ols, 'ensure_directories')
    @patch.object(reactive.ols, 'install_payload')
    @patch.object(hookenv, 'application_version_set')
    def test_handle_build_label_not_set(
        self, app_ver_set, install_payload, ensure_directories
    ):
        install_payload.side_effect = base.PayloadError('this is an error')
        self.fakes.charm_env.layer_config.update(
            {'ols': {'tarball_payload': True, 'symlink_switch_payload': False}}
        )
        bus.dispatch()
        self.fakes.status_set.assert_has_calls(
            [call('error', 'Failed to install payload: this is an error')]
        )
        install_payload.assert_called_once_with(
            base.payload_name(), base.code_dir(), 'root'
        )
        app_ver_set.assert_not_called()


class RenderTest(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.charm_dir = '/charm'
        self.fakes = self.useFixture(CharmEnvFixture(self.charm_dir))
        # below patching of juju-log is to silence it as it sends DEBUG logs to stdout
        ps = self.useFixture(FakeProcesses())
        juju_log_ok = Mock(return_value={'returncode': 0})
        ps.add(juju_log_ok, name='juju-log')

    def add_template(self, name, contents):
        template = '{}/templates/{}'.format(self.charm_dir, name)
        with open(template, 'w') as f:
            f.write(contents)

    def test_render_no_file(self):
        self.add_template('test.tmpl', '{{ foo }}')
        target = '/charm/file.cfg'
        context = {'foo': 'bar'}

        changed = render.render('test.tmpl', target, context)
        self.assertTrue(changed)
        with open(target) as f:
            output = f.read()
        self.assertEqual(output, 'bar')
        self.assertFalse(os.path.exists(target + '.tmp'))

    def test_render_changed(self):
        self.add_template('test.tmpl', '{{ foo }}')
        target = '/charm/file.cfg'
        context = {'foo': 'bar'}
        with open(target, 'w') as f:
            f.write('other')

        changed = render.render('test.tmpl', target, context)
        self.assertTrue(changed)
        with open(target) as f:
            output = f.read()
        self.assertEqual(output, 'bar')
        self.assertFalse(os.path.exists(target + '.tmp'))

    def test_render_not_changed(self):
        self.add_template('test.tmpl', '{{ foo }}')
        target = '/charm/file.cfg'
        context = {'foo': 'bar'}
        with open(target, 'w') as f:
            f.write('bar')
        mtime = os.stat(target).st_mtime

        changed = render.render('test.tmpl', target, context)
        self.assertFalse(changed)
        self.assertEqual(mtime, os.stat(target).st_mtime)
        self.assertFalse(os.path.exists(target + '.tmp'))

    @patch.object(render.distro, 'major_version')
    def test_render_logrotate_22_04(self, distro_major_version):
        distro_major_version.return_value = '22'
        context = {'log_path': '/logfile', 'logrotate_count': 1, 'group': 'ols'}
        render.render_logrotate('/srv/render-test', context)
        distro_major_version.assert_called_once()
        with open('/srv/render-test') as fd:
            content = fd.read()
        self.assertIn('systemctl kill -s HUP rsyslog', content)

    @patch.object(render.distro, 'major_version')
    def test_render_logrotate_20_04_or_older(self, distro_major_version):
        distro_major_version.return_value = '20'
        context = {'log_path': '/logfile', 'logrotate_count': 1, 'group': 'ols'}
        render.render_logrotate('/srv/render-test', context)
        distro_major_version.assert_called_once()
        with open('/srv/render-test') as fd:
            content = fd.read()
        self.assertIn('invoke-rc.d rsyslog rotate', content)
